import { Database, Change, ChangeTypes } from "./database";
import { resNull } from "../storage";
import nanoid = require("nanoid/generate");
import Logging from "@hibas123/nodelogging";
import * as MSGPack from "what-the-pack";

export const MP = MSGPack.initialize(2 ** 20);

const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

const { encode, decode } = MP;

interface ISubscribeOptions {
   existing: boolean;
}

export default class Query {
   /**
    * Returns true if the path only contains valid characters and false if it doesn't
    * @param path Path to be checked
    */
   private validatePath(path: string[]) {
      return path.every(e => (e.match(/[^a-zA-Z0-9_\-\<\>]/g) || []).length === 0);
   }

   constructor(private database: Database, private path: string[], private sender: string) {
      if (path.length > 10) {
         throw new Error("Path is to long. Path is only allowed to be 10 Layers deep!");
      }
      if (!this.validatePath(path)) {
         throw new Error("Path can only contain a-z A-Z 0-9  '-' '-' '<' and '>' ");
      }

      this.onChange = this.onChange.bind(this);
   }


   private async resolve(path: string[], create = false): Promise<{ collection: string, document: string, collectionKey: string }> {
      path = [...path]; // Create modifiable copy
      let collectionID: string = undefined;
      let documentKey = path.length % 2 === 0 ? path.pop() : undefined;
      let key = path.join("/");

      const lock = await this.database.collectionLocks.lock(key);

      try {
         collectionID = await this.database.collections.get(key).then(r => r.toString()).catch(resNull);
         if (!collectionID && create) {
            collectionID = nanoid(ALPHABET, 32);
            await this.database.collections.put(key, collectionID);
         }
      } finally {
         lock();
      }

      return {
         collection: collectionID,
         document: documentKey,
         collectionKey: key
      };
   }

   private getKey(collection: string, document?: string) {
      return `${collection || ""}/${document || ""}`;
   }

   private getDoc(collection: string, document: string) {
      return this.database.data.get(this.getKey(collection, document), { asBuffer: true }).then(res => decode(res as Buffer)).catch(resNull);
   }

   private sendChange(collection: string, document: string, type: ChangeTypes, data: any) {
      let change: Change = {
         type,
         document,
         data,
         sender: this.sender
      }

      let s = this.database.changes.get(this.getKey(collection, document))
      if (s)
         s.forEach(e => setImmediate(() => e(change)))
      s = this.database.changes.get(this.getKey(collection))
      if (s)
         s.forEach(e => setImmediate(() => e(change)))

   }

   public async get() {
      let { collection, document } = await this.resolve(this.path);

      if (!collection || !document) {
         return null;
      }
      const lock = await this.database.locks.lock(collection, document);
      return this.getDoc(collection, document).finally(() => lock())
   }

   public async keys() {
      let { collection, document } = await this.resolve(this.path);
      if (document)
         throw new Error("Keys only works on collections!");
      if (!collection)
         throw new Error("There must be a collection");


      let gt = Buffer.from(this.getKey(collection) + " ");
      gt[gt.length - 1] = 0;

      let lt = Buffer.alloc(gt.length);
      lt.set(gt);
      lt[gt.length - 1] = 0xFF;

      return new Promise<string[]>((yes, no) => {
         let keys = [];
         const stream = this.database.data.createKeyStream({
            gt,
            lt,
            keyAsBuffer: false
         })
         stream.on("data", (key: string) => {
            let s = key.split("/", 2);
            if (s.length > 1)
               keys.push(s[1]);
         });
         stream.on("end", () => yes(keys));
         stream.on("error", no);
      });
   }

   public async set(data: any, { merge = false }) {
      if (data === null)
         return this.delete();
      let { collection, document } = await this.resolve(this.path, true);
      if (!collection) {
         throw new Error("There must be a collection!")
      }

      if (!document) {
         throw new Error("There must be a document key!")
      }

      const lock = await this.database.locks.lock(collection, document);

      let isNew = await this.database.data.get(this.getKey(collection, document)).catch(resNull).then(e => e !== null);

      return this.database.data
         .put(this.getKey(collection, document), encode(data))
         .then(() => this.sendChange(collection, document, isNew ? "added" : "modified", data))
         .finally(() => lock())
   }

   public async push(value: any) {
      let id = nanoid(ALPHABET, 32);
      let q = new Query(this.database, [...this.path, id], this.sender);
      await q.set(value, {});
      return id;
   }

   public async delete() {
      let { collection, document } = await this.resolve(this.path);

      if (!collection) {
         throw new Error("There must be a collection!")
      }

      if (!document) {
         throw new Error("There must be a document key!")
      }


      const lock = await this.database.locks.lock(collection, document);

      return await this.database.data
         .del(`${collection}/${document}`)
         .then(() => this.sendChange(collection, document, "deleted", null))
         .finally(() => lock())
   }

   public async update(data: any) {
      //TODO: Implement
   }

   public async getCollections() {
      return new Promise<string[]>((yes, no) => {
         let keys = [];
         const stream = this.database.data.createKeyStream({ keyAsBuffer: false })
         stream.on("data", (key: string) => keys.push(key.split("/")));
         stream.on("end", () => yes(keys));
         stream.on("error", no);
      });
   }

   public async deleteCollection() {
      const { collection, document, collectionKey } = await this.resolve(this.path);

      if (document) {
         throw new Error("There can be no document defined on this operation");
      }

      let batch = this.database.data.batch();
      try {
         if (collection) {
            let keys = await this.keys();
            for (let key in keys) {
               batch.del(key);
            }
            await batch.write();
            batch = undefined;
            await this.database.collections.del(collectionKey);
         }
      } finally {
         if (batch)
            batch.clear();
      }
   }



   private subscription: {
      key: string;
      type: Set<ChangeTypes>;
      send: (data: Omit<Change, "sender">) => void;
   };
   async subscribe(send: (data: Omit<Change, "sender">) => void, type: ChangeTypes[] | undefined, _options: Partial<ISubscribeOptions>) {
      let options: ISubscribeOptions = {
         existing: true,
         ..._options
      }

      let { collection, document } = await this.resolve(this.path);

      let key = this.getKey(collection, document);
      let s = this.database.changes.get(key) || new Set();
      s.add(this.onChange)
      this.database.changes.set(key, s);

      this.subscription = {
         key,
         send,
         type: new Set(type || [])
      };

      Logging.debug("Existing?", options.existing)

      if (options.existing) {
         if (document) {
            send({
               document: document,
               type: "added",
               data: await this.get(),
            })
         } else {
            await this.keys().then(async documents => {
               for (let document of documents) {
                  const data = await this.getDoc(collection, document);
                  if (data)
                     send({
                        type: "added",
                        document,
                        data
                     })
               }
            })
         }
      }

      Logging.debug("Subscribed");
   }

   async onChange(change: Change) {
      const { type, send } = this.subscription;

      // Events from the sender are handled locally
      if (change.sender !== this.sender && type.has(change.type)) {
         let c = { ...change };
         delete c.sender;
         send(c)
      }
   }

   unsubscribe() {
      if (!this.subscription) {
         const { key } = this.subscription;
         let s = this.database.changes.get(key);
         if (s) {
            s.delete(this.onChange);
            if (s.size <= 0)
               this.database.changes.delete(key);
         }
         this.subscription = undefined;
      }
   }
}